"""This module defines functions to generate profiles with different shapes."""

from typing import Optional

import torch
from torch import Tensor

from ..functional import initialize_tensor
from ..planar_geometry import PlanarGeometry
from ..type_defs import Vector2

__all__ = ["checkerboard", "circle", "rectangle", "square"]


def checkerboard(
    shape: Vector2,
    tile_length: Vector2,
    num_tiles: Vector2,
    spacing: Optional[Vector2] = None,
    offset: Optional[Vector2] = None,
) -> Tensor:
    """
    Generates a checkerboard pattern.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        tile_length (Vector2): The side lengths of the checkerboard tiles.
        num_tiles (Vector2): Number of tiles along the planar dimensions.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Offset coordinates of the pattern. Default: `(0, 0)`.

    Returns:
        Tensor: The generated checkerboard pattern with internal padding.
    """
    x, y = PlanarGeometry(shape, 0, spacing, offset).meshgrid()
    tile_length = initialize_tensor("tile_length", tile_length, (2,), validate_positive=True, fill_value=True)
    num_tiles = initialize_tensor(
        "num_tiles", num_tiles, (2,), is_integer=True, validate_positive=True, fill_value=True
    )

    x_tile = (x + (tile_length[0] / 2 if num_tiles[0] % 2 == 1 else 0)) // tile_length[0]
    y_tile = (y + (tile_length[1] / 2 if num_tiles[1] % 2 == 1 else 0)) // tile_length[1]

    pattern = (1 + x_tile + y_tile) % 2
    pattern[x.abs() * 2 >= tile_length[0] * num_tiles[0]] = 0
    pattern[y.abs() * 2 >= tile_length[1] * num_tiles[1]] = 0

    return pattern


def circle(
    shape: Vector2, radius: float, spacing: Optional[Vector2] = None, offset: Optional[Vector2] = None
) -> Tensor:
    """
    Generates a circular profile.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        radius (float): The radius of the circle.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the profile. Default: `(0, 0)`.

    Returns:
        Tensor: The generated circular profile.
    """
    x, y = PlanarGeometry(shape, 0, spacing, offset).meshgrid()
    r = torch.sqrt(x**2 + y**2)
    return r <= radius  # type: ignore[attr-defined]


def rectangle(
    shape: Vector2, side: Vector2, spacing: Optional[Vector2] = None, offset: Optional[Vector2] = None
) -> Tensor:
    """
    Generates a rectangle profile.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        side (Vector2): The two side lengths of the rectangle.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the profile. Default: `(0, 0)`.

    Returns:
        Tensor: The generated rectangle profile.
    """
    x, y = PlanarGeometry(shape, 0, spacing, offset).meshgrid()
    side = initialize_tensor("side", side, (2,), validate_positive=True, fill_value=True)
    return (x.abs() <= side[0] / 2) & (y.abs() <= side[1] / 2)  # type: ignore


def square(
    shape: Vector2, side: float, spacing: Optional[Vector2] = None, offset: Optional[Vector2] = None
) -> Tensor:
    """
    Generates a square profile.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        side (float): The side length of the square.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the profile. Default: `(0, 0)`.

    Returns:
        Tensor: The generated square profile.
    """
    return rectangle(shape, (side, side), spacing, offset)



"""This module defines the Detector elements."""

from typing import Any, Optional

from torch import Tensor
from torch.nn.functional import linear

from ..fields import Field
from ..param import Param
from ..type_defs import Scalar, Vector2
from .elements import Element

__all__ = ["Detector", "IntensityDetector", "FieldDetector"]


class Detector(Element):
    r"""
    Detector element.

    Computes the power measured by the detector for each grid cell using the following equation:

    .. math::
        P_{i,j} = I_{i,j} \cdot \Delta A_{\text{cell}}

    where:
        - :math:`P_{i,j}` is the power measured in the grid cell at position :math:`(i, j)`,
        - :math:`I_{i,j}` is the intensity at position :math:`(i, j)`, and
        - :math:`\Delta A_{\text{cell}}` is the area of a single grid cell.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    def forward(self, field: Field) -> Tensor:
        """
        Calculates the power per cell area.

        Args:
            field (Field): The input field.

        Returns:
            Tensor: The power per cell area.
        """
        self.validate_field_geometry(field)
        return field.intensity() * self.cell_area()


class IntensityDetector(Element):
    r"""
    Intensity detector element.

    Computes the total weighted power measured by the detector using the following equation:

    .. math::
        P_c = \sum_{i, j} w_{c, i, j} \cdot I_{i, j} \cdot \Delta A_{\text{cell}}

    where:
        - :math:`P_c` is the total weighted power measured by the detector for channel :math:`c`,
        - :math:`w_{c, i, j}` is the weight matrix for channel :math:`c`,
        - :math:`I_{i, j}` is the intensity at position :math:`(i, j)`, and
        - :math:`\Delta A_{\text{cell}}` is the area of a single grid cell.

    .. note::
        This equation in its integral form is expressed as:

        .. math::
            P_c = \int \int w_c(x, y) \cdot I(x, y) \, dx \, dy

    Args:
        weight (Tensor): Weight matrix of shape (C, H, W).
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    _weight_is_complex = False

    def __init__(
        self,
        weight: Tensor,
        z: Scalar = 0,
        spacing: Optional[Vector2] = None,
        offset: Optional[Vector2] = None,
    ) -> None:
        self._validate_weight(weight)
        super().__init__(weight.shape[1:], z, spacing, offset)
        self.register_optics_property("weight", weight, is_complex=self._weight_is_complex)

    def forward(self, field: Field) -> Tensor:
        """
        Calculates the weighted power.

        Args:
            field (Field): The input field.

        Returns:
            Tensor: The weighted power.
        """
        self.validate_field_geometry(field)
        intensity_flat, weight_flat = field.intensity().flatten(-2), self.weight.flatten(-2)
        return linear(intensity_flat, weight_flat) * self.cell_area()  # pylint: disable=not-callable

    def visualize(self, *index: int, sum_weight: bool = False, **kwargs) -> Any:
        """
        Visualizes the detector output or the weight matrix.

        Args:
            *index (int): The index of the channel to visualize.
            sum_weight (bool): Whether to plot the sum of the weight matrix. Default: `False`.
            **kwargs: Additional keyword arguments for visualization.
        """
        data = self.weight.sum(dim=0) if sum_weight else self.weight
        return self._visualize(data, index, **kwargs)

    @staticmethod
    def _validate_weight(tensor):
        if not isinstance(tensor, (Tensor, Param)):
            raise TypeError(f"Expected weight to be a tensor, but got {type(tensor).__name__}")
        if tensor.dim() != 3:
            raise ValueError(f"Expected weight to be a 3D tensor, but got {tensor.dim()}D")


class FieldDetector(IntensityDetector):
    r"""
    Field detector element.

    Computes the total weighted power from field data using the inner product of the field with a weight
    tensor, based on the following equation:

    .. math::
        P_c = \left| \sum_{i, j} w_{c, i, j} \cdot \psi_{i, j} \cdot \Delta A_{\text{cell}} \right|^2

    where:
        - :math:`P_c` is the total weighted power measured by the detector for channel :math:`c`,
        - :math:`w_{c, i, j}` is the weight matrix for channel :math:`c`,
        - :math:`\psi_{i, j}` is the field at position :math:`(i, j)`, and
        - :math:`\Delta A_{\text{cell}}` is the area of a single grid cell.

    .. note::
        This equation in its integral form is expressed as:

        .. math::
            P_c = \left| \int \int w_c(x, y) \cdot \psi(x, y) \, dx \, dy \right|^2

    Args:
        weight (Tensor): Weight matrix of shape (C, H, W).
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.

    """

    _weight_is_complex = True

    def forward(self, field: Field) -> Tensor:
        """
        Calculates the weighted power from the field using the inner product with the weight matrix.

        Args:
            field (Field): The input field.

        Returns:
            Tensor: The weighted power after applying the inner product and calculating the magnitude squared.
        """
        self.validate_field_geometry(field)
        data_flat, weight_flat = field.data.flatten(-2), self.weight.flatten(-2)
        inner_prod = linear(data_flat, weight_flat) * self.cell_area()  # pylint: disable=not-callable
        return inner_prod.abs().square()



"""This module defines functions to generate grating profiles."""

import math
from typing import Optional

import torch

from ..planar_geometry import PlanarGeometry
from ..type_defs import Vector2

__all__ = ["blazed_grating", "sinusoidal_amplitude_grating", "sinusoidal_phase_grating"]


def blazed_grating(
    shape: Vector2,
    period: float,
    spacing: Optional[Vector2] = None,
    offset: Optional[Vector2] = None,
    theta: float = 0,
) -> torch.Tensor:
    r"""
    Generates a blazed grating profile.

    The blazed grating profile is defined by the following equation:

    .. math::
        \mathcal{M}(x, y) = \exp\left( i \frac{2\pi}{\Lambda} (x \cos(\theta) + y \sin(\theta)) \right)

    where:

    - :math:`\Lambda` is the period of the grating, and
    - :math:`\theta` is the angle of the grating.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        period (float): The grating period (distance between adjacent grooves).
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the beam. Default: `(0, 0)`.
        theta (float): The angle of the grating in radians. Default: `0`.


    Returns:
        Tensor: The generated transmission function.
    """

    x, y = PlanarGeometry(shape, 0, spacing, offset).meshgrid()
    phase = 2 * torch.pi * (x * math.cos(theta) + y * math.sin(theta)) / period
    return torch.exp(1j * phase)


def sinusoidal_amplitude_grating(
    shape: Vector2,
    m: float,
    period: float,
    spacing: Optional[Vector2] = None,
    offset: Optional[Vector2] = None,
    theta: float = 0,
) -> torch.Tensor:
    r"""
    Generates a sinusoidal amplitude grating profile.

    The sinusoidal amplitude grating profile is defined by the following equation:

    .. math::
        \mathcal{M}(x, y) = \frac{1}{2} + \frac{m}{2} \cos\left(2\pi \frac{x \cos(\theta)
        + y \sin(\theta)}{\Lambda}\right)

    where:

    - :math:`m` is the amplitude contrast, and
    - :math:`\Lambda` is the period of the grating.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        m (float): The amplitude contrast.
        period (float): The grating period (distance between adjacent grooves).
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the beam. Default: `(0, 0)`.
        theta (float): The angle of the grating in radians. Default: `0`.

    Returns:
        Tensor: The generated transmission function.
    """

    x, y = PlanarGeometry(shape, 0, spacing, offset).meshgrid()
    amplitude = 0.5 + m / 2 * torch.cos(2 * torch.pi * (x * math.cos(theta) + y * math.sin(theta)) / period)
    return amplitude


def sinusoidal_phase_grating(
    shape: Vector2,
    m: float,
    period: float,
    spacing: Optional[Vector2] = None,
    offset: Optional[Vector2] = None,
    theta: float = 0,
) -> torch.Tensor:
    r"""
    Generates a sinusoidal phase grating profile.

    The sinusoidal phase grating profile is defined by the following equation:

    .. math::
        \mathcal{M}(x, y) = \exp\left( i \frac{m}{2} \sin\left(2\pi \frac{x \cos(\theta)
        + y \sin(\theta)}{\Lambda}\right) \right)

    where:

    - :math:`m` is the phase contrast, and
    - :math:`\Lambda` is the period of the grating.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        m (float): The phase contrast.
        period (float): The grating period (distance between adjacent grooves).
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the beam. Default: `(0, 0)`.
        theta (float): The angle of the grating in radians. Default: `0`.

    Returns:
        Tensor: The generated transmission function.
    """

    x, y = PlanarGeometry(shape, 0, spacing, offset).meshgrid()
    phase = m / 2 * torch.sin(2 * torch.pi * (x * math.cos(theta) + y * math.sin(theta)) / period)
    return torch.exp(1j * phase)



import unittest

import torch
from torch import Tensor

from torchoptics import OpticsModule, Param


class TestOpticsModule(unittest.TestCase):

    def test_initialization(self):
        module = OpticsModule()
        self.assertIsInstance(module, OpticsModule)

    def test_register_property_as_parameter(self):
        module = OpticsModule()
        value = torch.tensor([1.0, 2.0, 3.0])
        module.register_optics_property("prop1", Param(value))
        self.assertTrue(hasattr(module, "prop1"))
        self.assertIsInstance(module.prop1, Tensor)
        self.assertTrue(module.prop1.requires_grad)
        self.assertIn("prop1", dict(module.named_parameters()))

    def test_register_property_as_buffer(self):
        module = OpticsModule()
        value = torch.tensor([1.0, 2.0, 3.0])
        module.register_optics_property("prop1", value)
        self.assertTrue(hasattr(module, "prop1"))
        self.assertIsInstance(module.prop1, Tensor)
        self.assertFalse(module.prop1.requires_grad)
        self.assertIn("prop1", dict(module.named_buffers()))

    def test_set_property(self):
        module = OpticsModule()
        initial_value = torch.tensor([1.0, 2.0, 3.0])
        module.register_optics_property("prop1", initial_value)
        new_value = torch.tensor([4.0, 5.0, 6.0])
        module.set_optics_property("prop1", new_value)
        self.assertTrue(torch.equal(module.prop1, new_value))

    def test_property_shape_validation(self):
        module = OpticsModule()
        initial_value = torch.tensor([1.0, 2.0, 3.0])
        module.register_optics_property("prop1", initial_value)
        new_value = torch.tensor([[1.0, 2.0], [3.0, 4.0]])
        with self.assertRaises(ValueError):
            module.set_optics_property("prop1", new_value)

    def test_complex_property_registration(self):
        module = OpticsModule()
        value = torch.tensor([1.0 + 1.0j, 2.0 + 2.0j, 3.0 + 3.0j])
        module.register_optics_property("prop1", value, is_complex=True)
        self.assertTrue(module.prop1.is_complex())

    def test_validate_positive(self):
        module = OpticsModule()
        value = torch.tensor([1.0, 2.0, 3.0])
        module.register_optics_property("prop1", value, validate_positive=True)
        invalid_value = torch.tensor([-1.0, 2.0, 3.0])
        with self.assertRaises(ValueError):
            module.set_optics_property("prop1", invalid_value)

    def test_fill_value(self):
        module = OpticsModule()
        value = 2.0
        module.register_optics_property("prop1", value, shape=(3,), fill_value=True)
        expected_tensor = torch.tensor([2.0, 2.0, 2.0], dtype=torch.double)
        self.assertTrue(torch.equal(module.prop1, expected_tensor))

    def test_register_property_before_init(self):
        module = OpticsModule()
        module._initialized = False
        with self.assertRaises(AttributeError):
            module.register_optics_property("prop1", torch.tensor([1.0]))

    def test_register_property_from_sequence(self):
        module = OpticsModule()
        value = [1.0, 2.0, 3.0]
        module.register_optics_property("prop1", value, shape=(3,))
        self.assertTrue(hasattr(module, "prop1"))
        self.assertIsInstance(module.prop1, Tensor)
        expected_tensor = torch.tensor([1.0, 2.0, 3.0], dtype=torch.double)
        self.assertTrue(torch.equal(module.prop1, expected_tensor))
        self.assertFalse(module.prop1.requires_grad)
        self.assertIn("prop1", dict(module.named_buffers()))

    def test_register_trainable_property_from_sequence(self):
        module = OpticsModule()
        value = [1.0, 2.0, 3.0]
        module.register_optics_property("prop1", Param(value), shape=(3,))
        self.assertTrue(hasattr(module, "prop1"))
        self.assertIsInstance(module.prop1, Tensor)
        expected_tensor = torch.tensor([1.0, 2.0, 3.0], dtype=torch.double)
        self.assertTrue(torch.equal(module.prop1, expected_tensor))
        self.assertTrue(module.prop1.requires_grad)
        self.assertIn("prop1", dict(module.named_parameters()))

    def test_register_property_with_none_shape(self):
        module = OpticsModule()
        value = torch.tensor([1.0, 2.0, 3.0])
        module.register_optics_property("prop1", value)
        self.assertTrue(hasattr(module, "prop1"))
        self.assertIsInstance(module.prop1, Tensor)
        self.assertFalse(module.prop1.requires_grad)
        self.assertEqual(module.prop1.shape, (3,))

    def test_register_trainable_property_with_none_shape(self):
        module = OpticsModule()
        value = torch.tensor([1.0, 2.0, 3.0])
        module.register_optics_property("prop1", Param(value))
        self.assertTrue(hasattr(module, "prop1"))
        self.assertIsInstance(module.prop1, Tensor)
        self.assertTrue(module.prop1.requires_grad)
        self.assertEqual(module.prop1.shape, (3,))

    def test_set_property_via_setattr(self):
        module = OpticsModule()
        initial_value = [1.0, 2.0, 3.0]
        module.register_optics_property("prop1", initial_value, shape=(3,))
        new_value = [4.0, 5.0, 6.0]
        module.prop1 = new_value  # Using setattr
        expected_tensor = torch.tensor(new_value, dtype=torch.double)
        self.assertTrue(torch.equal(module.prop1, expected_tensor))

    def test_set_property_via_setattr_tensor(self):
        module = OpticsModule()
        initial_value = [1.0, 2.0, 3.0]
        module.register_optics_property("prop1", initial_value, shape=(3,))
        new_value = torch.tensor([4.0, 5.0, 6.0])
        module.prop1 = new_value  # Using setattr
        self.assertTrue(torch.equal(module.prop1, new_value))

    def test_set_trainable_property_via_setattr(self):
        module = OpticsModule()
        initial_value = [1.0, 2.0, 3.0]
        module.register_optics_property("prop1", Param(initial_value), shape=(3,))
        new_value = [4.0, 5.0, 6.0]
        # Prevents RuntimeError: a leaf Variable that requires grad is being used in an in-place operation.
        with torch.no_grad():
            module.prop1 = new_value  # Using setattr
        expected_tensor = torch.tensor(new_value, dtype=torch.double)
        self.assertTrue(torch.equal(module.prop1, expected_tensor))

    def test_raise_errors(self):
        with self.assertRaises(ValueError):
            OpticsModule().register_optics_property("prop1", 1.0)
        with self.assertRaises(AttributeError):
            OpticsModule().set_optics_property("unregistered_prop", 1.0)
        with self.assertRaises(ValueError):
            OpticsModule().register_optics_property("complex_prop", 1.0, is_complex=True)


if __name__ == "__main__":
    unittest.main()



"""This module defines functions to generate airy and sinc profiles."""

from typing import Optional

import torch
from torch import Tensor
from torch.special import bessel_j1  # Bessel function of the first kind

from ..functional import initialize_tensor
from ..planar_geometry import PlanarGeometry
from ..type_defs import Vector2

__all__ = ["airy", "sinc"]


def airy(
    shape: Vector2, scale: float, spacing: Optional[Vector2] = None, offset: Optional[Vector2] = None
) -> Tensor:
    r"""
    Generates an Airy pattern profile.

    The Airy pattern is defined by the following equation:

    .. math::
        \psi(r) = \left( \frac{2 J_1(\frac{r}{a})}{\frac{r}{a}} \right)^2

    where:

    - :math:`r` is the radial distance from the center of the diffraction pattern.
    - :math:`J_1` is the Bessel function of the first kind.
    - :math:`a` is the scaling factor that determines the size of the Airy disk.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        scale (float): A scaling factor that determines the size of the Airy disk.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default.
        offset (Optional[Vector2]): Center coordinates of the profile. Default: `(0, 0)`.

    Returns:
        Tensor: The generated Airy profile.
    """
    x, y = PlanarGeometry(shape, 0, spacing, offset).meshgrid()
    r = torch.sqrt(x**2 + y**2)
    scaled_r = r / scale
    airy_pattern = (2 * bessel_j1(scaled_r) / (scaled_r)) ** 2  # pylint: disable=not-callable
    airy_pattern[r == 0] = 1.0  # Handle the value at r = 0
    return airy_pattern


def sinc(
    shape: Vector2, scale: Vector2, spacing: Optional[Vector2] = None, offset: Optional[Vector2] = None
) -> Tensor:
    r"""
    Generates a sinc profile.

    The sinc profile is defined by the following equation:

    .. math::
        \psi(x, y) = \text{sinc}\left(\frac{x}{a}\right) \cdot \text{sinc}\left(\frac{y}{b}\right)

    where:

    - :math:`\text{sinc}(x) = \frac{\sin(\pi x)}{\pi x}` is the normalized sinc function.
    - :math:`a` and :math:`b` are the scaling factors along the x and y dimensions, respectively.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        scale (Vector2): The two scaling factors (widths) of the sinc function in the x and y directions.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default.
        offset (Optional[Vector2]): Center coordinates of the profile. Default: `(0, 0)`.

    Returns:
        Tensor: The generated sinc profile.
    """
    x, y = PlanarGeometry(shape, 0, spacing, offset).meshgrid()
    scale = initialize_tensor("scale", scale, (2,), fill_value=True, validate_positive=True)
    sinc_pattern = torch.sinc(x / scale[0]) * torch.sinc(y / scale[1])
    return sinc_pattern



"""This module defines the waveplate element."""

from typing import Optional

import torch
from torch import Tensor

from ..type_defs import Scalar, Vector2
from .elements import PolarizedModulationElement

__all__ = ["Waveplate", "QuarterWaveplate", "HalfWaveplate"]


class Waveplate(PolarizedModulationElement):
    r"""
    Waveplate element.

    The waveplate is described by the following polarization matrix:

    .. math::
        J = 
        \begin{bmatrix}
            \cos^2(\theta) + e^{i \phi} \sin^2(\theta) & (1 - e^{i \phi}) \cos(\theta) \sin(\theta) & 0 \\
            (1 - e^{i \phi}) \cos(\theta) \sin(\theta) & \sin^2(\theta) + e^{i \phi} \cos^2(\theta) & 0 \\
            0 & 0 & 1
        \end{bmatrix}

    where 
        - :math:`\theta` is the fast axis angle, and 
        - :math:`\phi` is the phase delay between the fast and slow axes.

    .. note::
        A quarter waveplate (QWP) is obtained by setting :math:`\phi = \pi/2`. 

        A half waveplate (HWP) is obtained by setting :math:`\phi = \pi`. 

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        z (Scalar): Position along the z-axis. Default: `0`.
        phi (Scalar): Phase delay of the waveplate.
        theta (Scalar): Fast axis angle of the waveplate.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    def __init__(
        self,
        shape: Vector2,
        phi: Scalar,
        theta: Scalar,
        z: Scalar = 0,
        spacing: Optional[Vector2] = None,
        offset: Optional[Vector2] = None,
    ) -> None:
        super().__init__(shape, z, spacing, offset)
        self.register_optics_property("phi", phi, ())
        self.register_optics_property("theta", theta, ())

    @property
    def polarized_modulation_profile(self) -> Tensor:
        """Return the polarized modulation profile."""
        tensor = torch.zeros(3, 3, *self.shape, dtype=torch.cdouble, device=next(self.buffers()).device)
        tensor[0, 0] = torch.cos(self.theta) ** 2 + torch.exp(1j * self.phi) * torch.sin(self.theta) ** 2
        tensor[0, 1] = (1 - torch.exp(1j * self.phi)) * torch.cos(self.theta) * torch.sin(self.theta)
        tensor[1, 0] = (1 - torch.exp(1j * self.phi)) * torch.cos(self.theta) * torch.sin(self.theta)
        tensor[1, 1] = torch.sin(self.theta) ** 2 + torch.exp(1j * self.phi) * torch.cos(self.theta) ** 2
        tensor[2, 2] = 1
        return tensor


class QuarterWaveplate(PolarizedModulationElement):
    r"""
    Quarter Waveplate element.

    The quarter waveplate is described by the following polarization matrix:

    .. math::
        J = 
        \begin{bmatrix}
            \cos^2(\theta) + i \sin^2(\theta) & (1 - i) \cos(\theta) \sin(\theta) & 0 \\
            (1 - i) \cos(\theta) \sin(\theta) & \sin^2(\theta) + i \cos^2(\theta) & 0 \\
            0 & 0 & 1
        \end{bmatrix}

    where 
        - :math:`\theta` is the fast axis angle.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        z (Scalar): Position along the z-axis. Default: `0`.
        theta (Scalar): Fast axis angle of the waveplate.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    def __init__(
        self,
        shape: Vector2,
        theta: Scalar,
        z: Scalar = 0,
        spacing: Optional[Vector2] = None,
        offset: Optional[Vector2] = None,
    ) -> None:
        super().__init__(shape, z, spacing, offset)
        self.register_optics_property("theta", theta, ())

    @property
    def polarized_modulation_profile(self) -> Tensor:
        """Return the polarized modulation profile."""
        tensor = torch.zeros(3, 3, *self.shape, dtype=torch.cdouble, device=next(self.buffers()).device)
        tensor[0, 0] = torch.cos(self.theta) ** 2 + 1j * torch.sin(self.theta) ** 2
        tensor[0, 1] = (1 - 1j) * torch.cos(self.theta) * torch.sin(self.theta)
        tensor[1, 0] = (1 - 1j) * torch.cos(self.theta) * torch.sin(self.theta)
        tensor[1, 1] = torch.sin(self.theta) ** 2 + 1j * torch.cos(self.theta) ** 2
        tensor[2, 2] = 1
        return tensor


class HalfWaveplate(PolarizedModulationElement):
    r"""
    Half Waveplate element.

    The half waveplate is described by the following polarization matrix:

    .. math::
        J = 
        \begin{bmatrix}
            \cos^2(\theta) - \sin^2(\theta) & 2 \cos(\theta) \sin(\theta) & 0 \\
            2 \cos(\theta) \sin(\theta) & \sin^2(\theta) - \cos^2(\theta) & 0 \\
            0 & 0 & 1
        \end{bmatrix}

    where 
        - :math:`\theta` is the fast axis angle.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        z (Scalar): Position along the z-axis. Default: `0`.
        theta (Scalar): Fast axis angle of the waveplate.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    def __init__(
        self,
        shape: Vector2,
        theta: Scalar,
        z: Scalar = 0,
        spacing: Optional[Vector2] = None,
        offset: Optional[Vector2] = None,
    ) -> None:
        super().__init__(shape, z, spacing, offset)
        self.register_optics_property("theta", theta, ())

    @property
    def polarized_modulation_profile(self) -> Tensor:
        """Return the polarized modulation profile."""
        tensor = torch.zeros(3, 3, *self.shape, dtype=torch.cdouble, device=next(self.buffers()).device)
        tensor[0, 0] = torch.cos(self.theta) ** 2 - torch.sin(self.theta) ** 2
        tensor[0, 1] = 2 * torch.cos(self.theta) * torch.sin(self.theta)
        tensor[1, 0] = 2 * torch.cos(self.theta) * torch.sin(self.theta)
        tensor[1, 1] = torch.sin(self.theta) ** 2 - torch.cos(self.theta) ** 2
        tensor[2, 2] = 1
        return tensor



"""This module defines functions to generate a thin lens profile."""

from typing import Optional

import torch

from ..config import get_default_wavelength
from ..planar_geometry import PlanarGeometry
from ..type_defs import Scalar, Vector2

__all__ = ["lens"]


def lens(
    shape: Vector2,
    focal_length: Scalar,
    z: Scalar = 0,
    wavelength: Optional[Scalar] = None,
    spacing: Optional[Vector2] = None,
    offset: Optional[Vector2] = None,
    is_circular_lens: bool = True,
):
    r"""
    Generates a thin lens profile.

    The thin lens profile is defined by the following equation:

    .. math::
        \mathcal{M}(x, y) = \exp\left(-i \frac{\pi}{\lambda f} (x^2 + y^2) \right)

    where:
        - :math:`\lambda` is the wavelength of the light, and
        - :math:`f` is the focal length of the lens.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        z (Scalar): Position along the z-axis. Default: `0`.
        focal_length (Scalar): Focal length of the lens.
        wavelength (Optional[Scalar]): Wavelength used for lens operation. Default: if `None`, uses a
            global default (see :meth:`torchoptics.config.set_default_wavelength()`).
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
        is_circular_lens (bool): If `True`, the lens is circular and the phase profile is set to zero outside
            the lens diameter, otherwise lens is square. Default: `True`.
    """
    wavelength = wavelength if wavelength is not None else get_default_wavelength()

    planar_geometry = PlanarGeometry(shape, z, spacing, offset)
    x, y = planar_geometry.meshgrid()
    radial_dist = torch.sqrt(x**2 + y**2)
    phase_profile = torch.exp(-1j * torch.pi / (wavelength * focal_length) * radial_dist**2)  # type: ignore

    if is_circular_lens:
        lens_diameter = min(planar_geometry.length(use_grid_points=True))
        mask = radial_dist > lens_diameter / 2
        phase_profile[mask] = 0
    return phase_profile



"""This module defines methods for getting and setting default values for spacing and wavelength."""

from typing import Optional

from torch import Tensor

from .functional import initialize_tensor
from .type_defs import Scalar, Vector2

__all__ = ["get_default_spacing", "set_default_spacing", "get_default_wavelength", "set_default_wavelength"]


class Config:
    """Global configuration values for torchoptics."""

    spacing: Optional[Tensor] = None
    wavelength: Optional[Tensor] = None


def get_default_spacing() -> Tensor:
    """Gets the current default ``spacing`` value."""
    default_spacing = Config.spacing
    if default_spacing is None:
        raise ValueError("Default spacing is not set.")
    return default_spacing


def set_default_spacing(value: Vector2) -> None:
    """
    Sets the default ``spacing`` value.

    Args:
        value (Vector2): The default spacing.

    Example:
        >>> torchoptics.set_default_spacing((10e-6, 10e-6))
    """
    Config.spacing = initialize_tensor("spacing", value, (2,), validate_positive=True, fill_value=True)


def get_default_wavelength() -> Tensor:
    """Gets the current default ``wavelength`` value."""
    default_wavelength = Config.wavelength
    if default_wavelength is None:
        raise ValueError("Default wavelength is not set.")
    return default_wavelength


def set_default_wavelength(value: Scalar) -> None:
    """
    Sets the default ``wavelength`` value.

    Args:
        value (Scalar): The default wavelength.

    Example:
        >>> torchoptics.set_default_wavelength(700e-6)
    """
    Config.wavelength = initialize_tensor("wavelength", value, (), validate_positive=True)



"""This module defines the System class."""

from typing import Optional

from torch.nn import Module, ModuleList

from .elements import Element, IdentityElement, ModulationElement, PolarizedModulationElement
from .fields import Field
from .planar_geometry import PlanarGeometry
from .type_defs import Scalar, Vector2

__all__ = ["System"]


class System(Module):
    """
    System of optical elements similar to the :class:`torch.nn.Sequential` module.

    The system is defined by a sequence of optical elements which are sorted by their ``z`` position.
    The :meth:`forward()` method accepts a :class:`Field` object as input. The field is
    propagated to the first element in the system which processes it using its ``forward()`` method.
    The field is then propagated to the next element in the system and so on, finally returning the
    field after it has been processed by the last element in the system.

    Example:
        Initialize a 4f optical system with two lenses::

            import torch
            import torchoptics
            from torchoptics import Field, System
            from torchoptics.elements import Lens

            # Set simulation properties
            shape = 1000  # Number of grid points in each dimension
            spacing = 10e-6  # Spacing between grid points (m)
            wavelength = 700e-9  # Field wavelength (m)
            focal_length = 200e-3  # Lens focal length (m)

            # Determine device
            device = "cuda" if torch.cuda.is_available() else "cpu"

            # Configure torchoptics default properties
            torchoptics.set_default_spacing(spacing)
            torchoptics.set_default_wavelength(wavelength)

            # Define 4f optical system with two lenses
            system = System(
                Lens(shape, focal_length, z=1 * focal_length),
                Lens(shape, focal_length, z=3 * focal_length),
            ).to(device)

    Args:
        *elements (Element): Optical elements in the system.
    """

    def __init__(self, *elements: Element) -> None:
        super().__init__()
        self.elements = ModuleList(elements)

    def __iter__(self):
        return iter(self.elements)

    def __getitem__(self, index):
        return self.elements[index]

    def __len__(self):
        return len(self.elements)

    def forward(self, field: Field) -> Field:
        """
        Propagates the field through the system.

        Args:
            field (Field): Input field.

        Returns:
            Field: Output field after propagating through the system."""
        return self._forward(field)

    def measure(
        self,
        field: Field,
        shape: Vector2,
        z: Scalar,
        spacing: Optional[Vector2] = None,
        offset: Optional[Vector2] = None,
    ) -> Field:
        """
        Propagates the field through the system to a plane defined by the input parameters.

        Args:
            field (Field): Input field.
            shape (Vector2): Number of grid points along the planar dimensions.
            z (Scalar): Position along the z-axis.
            spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default:
                if `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
            offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.

        Returns:
            Field: Output field after propagating to the plane.
        """
        output_element = IdentityElement(shape, z, spacing, offset).to(field.data.device)
        return self._forward(field, output_element)

    def measure_at_z(self, field: Field, z: Scalar) -> Field:
        """
        Propagates the field through the system to a plane at a specific z position.

        The plane has the same ``shape``, ``spacing``, and ``offset`` as the input field.

        Args:
            field (Field): Input field.
            z (Scalar): Position along the z-axis.

        Returns:
            Field: Output field after propagating to the plane.
        """
        return self.measure(field, field.shape, z, field.spacing, field.offset)

    def measure_at_plane(self, field: Field, plane: PlanarGeometry) -> Field:
        """
        Propagates the field through the system to a plane defined by a :class:`PlanarGeometry` object.

        Args:
            field (Field): Input field.
            plane (PlanarGeometry): Plane geometry.

        Returns:
            Field: Output field after propagating to the plane.
        """
        return self.measure(field, plane.shape, plane.z, plane.spacing, plane.offset)

    def sorted_elements(self) -> tuple[Element, ...]:
        """Returns the elements sorted by their z position."""
        return tuple(sorted(self.elements, key=lambda element: element.z))

    def elements_in_field_path(self, field: Field, output_element: Optional[Element]) -> tuple[Element, ...]:
        """
        Returns the elements along the field path.

        Args:
            field (Field): Input field.
            output_element (Optional[Element]): Output element.

        Returns:
            tuple[Element]: Elements along the field path.
        """
        elements_in_path = [element for element in self.sorted_elements() if field.z <= element.z]
        if output_element:
            elements_in_path = [element for element in elements_in_path if element.z <= output_element.z]
            if elements_in_path and isinstance(elements_in_path[-1], IdentityElement):
                elements_in_path.pop()
            elements_in_path.append(output_element)

        self._validate_elements_in_field_path(field, elements_in_path)
        return tuple(elements_in_path)

    def _validate_elements_in_field_path(self, field, elements_in_path):
        if not elements_in_path:
            raise ValueError("Expected system to contain at least one element.")
        if elements_in_path[-1].z < field.z:
            raise ValueError(
                f"Expected last element z ({elements_in_path[-1].z}) to be greater than field z ({field.z})."
            )
        if not all(
            isinstance(element, (ModulationElement, PolarizedModulationElement))
            for element in elements_in_path[:-1]
        ):
            raise TypeError(
                "Expected all elements in field path, except for the last one, to be type "
                "BaseModulationElement or BasePolarizedModulationElement."
            )
        if not isinstance(elements_in_path[-1], Element):
            raise TypeError("Expected the last element to be type Element.")

    def _forward(self, field: Field, output_element: Optional[Element] = None) -> Field:
        """Propagates the field through the system to the output element if provided."""
        for element in self.elements_in_field_path(field, output_element):
            field = field.propagate_to_plane(element)
            field = element(field)
        return field



"""This module defines the OpticsModule class."""

from typing import Any, Optional, Sequence

from torch import Tensor
from torch.nn import Module, Parameter

from .functional import initialize_tensor
from .param import Param

__all__ = ["OpticsModule"]


class OpticsModule(Module):  # pylint: disable=abstract-method
    """
    Base class for all optics modules.

    This class facilitates the registration of tensors, representing optics-related properties, as either
    PyTorch parameters or buffers. These properties are validated and registered using
    :meth:`register_optics_property()`::

        from torchoptics import OpticsModule, Param

        class MyOpticsModule(OpticsModule):
            def __init__(self, trainable_property, non_trainable_property):
                super().__init__()
                self.register_optics_property("trainable_property", Param(trainable_property), shape=())
                self.register_optics_property("non_trainable_property", non_trainable_property, shape=())

    Once the properties are registered, they can be updated using :meth:`set_optics_property()`.

    .. note::
        :meth:`__setattr__()` is overridden to call :meth:`set_optics_property()` when setting the value of an
        optics property.

    """

    _initialized = False

    def __init__(self) -> None:
        super().__init__()
        self._optics_property_configs: dict[str, dict] = {}
        self._initialized = True

    def __setattr__(self, name: str, value: Any) -> None:
        """
        Sets the attribute of the module.

        If the attribute name is a registered optics property, the optics property value is set using the
        :meth:`set_optics_property()` method. Otherwise, the attribute is set normally.

        Args:
            name (str): The name of the attribute.
            value (Any): The value to set for the attribute.
        """
        if self._initialized and name in self._optics_property_configs:
            self.set_optics_property(name, value)
        else:
            super().__setattr__(name, value)

    def register_optics_property(
        self,
        name: str,
        value: Any,
        shape: Optional[Sequence] = None,
        is_complex: bool = False,
        validate_positive: bool = False,
        validate_non_negative: bool = False,
        fill_value: bool = False,
    ) -> None:
        """
        Registers an optics property as a PyTorch parameter or buffer.

        Args:
            name (str): Name of the optics property.
            value (Any): Initial value of the property.
            shape (Optional[Sequence]): Shape of the property tensor. Required if value is not a tensor.
                Default: `None`.
            is_complex (bool): Whether the property tensor is complex. Default: `False`.
            validate_positive (bool): Whether to validate that the property tensor contains only positive
                values. Default: `False`.
            validate_non_negative (bool): Whether to validate that the property tensor contains only
                non-negative. Default: `False`.
            fill_value (bool): Whether to fill the tensor with the initial value if it is a scalar.
                Default: `False`.

        Raises:
            AttributeError: If called before the class is initialized.
            ValueError: If shape is not provided and value is not a tensor, or if the value does not match the
                property's conditions.
        """
        if not self._initialized:
            raise AttributeError("Cannot register optics property before __init__() call.")
        is_param = isinstance(value, Param)
        value = value.data if is_param else value
        if shape is None:
            if not isinstance(value, Tensor):
                raise ValueError(f"shape must be provided if {name} is not a tensor.")
            shape = value.shape

        property_config = {
            "name": name,
            "shape": shape,
            "is_complex": is_complex,
            "validate_positive": validate_positive,
            "validate_non_negative": validate_non_negative,
            "fill_value": fill_value,
        }
        self._optics_property_configs[name] = property_config

        tensor = initialize_tensor(value=value, **property_config)  # type: ignore
        if is_param:
            self.register_parameter(name, Parameter(tensor))
        else:
            self.register_buffer(name, tensor)

    def set_optics_property(self, name: str, value: Any) -> None:
        """
        Sets the value of an existing optics property.

        Args:
            name (str): Name of the optics property.
            value (Any): New value of the property.

        Raises:
            AttributeError: If the property is not registered.
            ValueError: If the value does not match the property's conditions.
        """
        if self._initialized and name in self._optics_property_configs:
            updated_tensor = initialize_tensor(value=value, **self._optics_property_configs[name])
            attr_tensor = getattr(self, name)
            attr_tensor.copy_(updated_tensor)
        else:
            raise AttributeError(f"Cannot set unknown optics property: {name}.")



"""This module defines functions to propagate Field objects."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import torch

from ..functional import plane_sample
from ..planar_geometry import PlanarGeometry
from ..type_defs import Scalar, Vector2
from .angular_spectrum_method import asm_propagation
from .direct_integration_method import dim_propagation

if TYPE_CHECKING:
    from ..fields import Field

__all__ = [
    "propagator",
    "get_propagation_plane",
    "is_dim_propagation",
    "calculate_min_dim_propagation_distance",
]


def propagator(
    field: Field,
    shape: Vector2,
    z: Scalar,
    spacing: Optional[Vector2],
    offset: Optional[Vector2],
) -> Field:
    """
    Propagates the field through free-space to a plane defined by the input parameters.

    First, the field is propagated to the plane determined by :meth:`get_propagation_plane()`. This propagated
    field is then interpolated using :func:`torchoptics.functional.plane_sample()` to match the geometry of
    the output plane.

    Args:
        field (Field): Input field.
        shape (Vector2): Number of grid points along the planar dimensions.
        z (Scalar): Position along the z-axis.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.

    Returns:
        Field: Output field after propagating to the plane.
    """
    output_plane = PlanarGeometry(shape, z, spacing, offset).to(field.data.device)

    if output_plane.z != field.z:  # Propagate to output plane z
        propagation_plane = get_propagation_plane(field, output_plane)
        propagation_func = (
            dim_propagation if is_dim_propagation(field, propagation_plane) else asm_propagation
        )
        field = propagation_func(field, propagation_plane)

    if not output_plane.is_same_geometry(field):  # Interpolate to output plane geometry
        transformed_data = plane_sample(field.data, field, output_plane, field.interpolation_mode)
        field = field.copy(data=transformed_data, spacing=output_plane.spacing, offset=output_plane.offset)

    return field


def get_propagation_plane(field: Field, output_plane: PlanarGeometry) -> PlanarGeometry:
    r"""
    Creates a propagation plane that is equal to or slightly larger than the specified output plane.

    The propagation plane adopts the same ``spacing`` as the ``field``, and retains the same ``z`` and
    ``offset`` values as the ``output_plane``.

    The length of the propagation plane must satisfy the inequality:

    .. math::
        \text{propagation plane length} \geq \text{output plane length}.

    This can be expressed as:

    .. math::
        (N_{{\text{prop}}} - 1) \cdot \Delta_{{\text{prop}}} \geq (N_{{\text{out}}} - 1) \cdot
        \Delta_{{\text{out}}}.

    Therefore, the number of grid points in the propagation plane, :math:`N_{{\text{prop}}}`, must be:

    .. math::
        N_{{\text{prop}}} \geq \left [ \frac{{(N_{{\text{out}}} - 1)
        \cdot \Delta_{{\text{out}}}}}{{\Delta_{{\text{prop}}}}} \right ] + 1,

    where:

    - :math:`N_{{\text{prop}}}` is the number of grid points (``shape``) in the propagation plane.
    - :math:`N_{{\text{out}}}` is the number of grid points (``shape``) in the output plane.
    - :math:`\Delta_{{\text{prop}}}` is the spacing in the propagation plane.
    - :math:`\Delta_{{\text{out}}}` is the spacing in the output plane.
    """
    spacing_ratio = output_plane.spacing / field.spacing
    output_plane_shape = torch.tensor(output_plane.shape, device=spacing_ratio.device)
    propagation_shape = torch.ceil((output_plane_shape - 1) * spacing_ratio).int() + 1
    return PlanarGeometry(propagation_shape, output_plane.z, field.spacing, output_plane.offset)


def is_dim_propagation(field: Field, propagation_plane: PlanarGeometry) -> bool:
    """
    Returns whether propagation using DIM should be used.

    Returns `True` if :attr:`field.propagation_method` is `"DIM"` or `"DIM_FRESNEL"`.

    Returns `False` if :attr:`field.propagation_method` is `"ASM"` or `"ASM_FRESNEL"`.

    If :attr:`field.propagation_method` is `"auto"`, the propagation method is determined based on the
    condition set in :func:`calculate_min_dim_propagation_distance`.
    """
    if field.propagation_method in ("DIM", "DIM_FRESNEL"):
        return True
    if field.propagation_method in ("ASM", "ASM_FRESNEL"):
        return False
    abs_propagation_distance = abs(propagation_plane.z - field.z)
    return calculate_min_dim_propagation_distance(field) < abs_propagation_distance


def calculate_min_dim_propagation_distance(field: Field) -> float:
    r"""
    Calculates the minimum distance required for accurate simulation using DIM.

    The minimum distance is calculated using the criteria from D. Voelz's textbook "Computational Fourier
    Optics: A MATLAB Tutorial" (2011):

    .. math::
        z \geq L \Delta / \lambda,

    where:

    - :math:`z` is the propagation distance.
    - :math:`L` is the maximum length of the field.
    - :math:`\Delta` is the spacing of the field.
    - :math:`\lambda` is the wavelength of the field.
    """
    return float((field.length() * field.spacing).max() / field.wavelength)



"""This module defines type aliases for the torchoptics package."""

from typing import Sequence, Union

from torch import Tensor

from .param import Param

__all__ = ["Scalar", "Vector2"]

Scalar = Union[int, float, bool, complex, bytes, Tensor, Param]
Vector2 = Union[int, float, bool, complex, bytes, Tensor, Param, list, tuple, Sequence]



"""This module defines the polarizer elements."""

from typing import Optional

import torch
from torch import Tensor

from ..type_defs import Scalar, Vector2
from .elements import PolarizedModulationElement

__all__ = ["LinearPolarizer", "LeftCircularPolarizer", "RightCircularPolarizer"]


class LinearPolarizer(PolarizedModulationElement):
    r"""
    Linear polarizer element.

    The linear polarizer is described by the following polarization matrix:

    .. math::
        J = 
        \begin{bmatrix}
            \cos^2(\theta) & \cos(\theta) \sin(\theta) & 0 \\
            \cos(\theta) \sin(\theta) & \sin^2(\theta) & 0 \\
            0 & 0 & 1
        \end{bmatrix}

    where :math:`\theta` is the transmission axis of the polarizer.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        z (Scalar): Position along the z-axis. Default: `0`.
        theta (Scalar): Transmission axis of the polarizer.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    def __init__(
        self,
        shape: Vector2,
        theta: Scalar,
        z: Scalar = 0,
        spacing: Optional[Vector2] = None,
        offset: Optional[Vector2] = None,
    ) -> None:
        super().__init__(shape, z, spacing, offset)
        self.register_optics_property("theta", theta, ())

    @property
    def polarized_modulation_profile(self) -> Tensor:
        """Return the polarized modulation profile."""
        tensor = torch.zeros(3, 3, *self.shape, dtype=torch.cdouble, device=next(self.buffers()).device)
        tensor[0, 0] = torch.cos(self.theta) ** 2
        tensor[0, 1] = torch.cos(self.theta) * torch.sin(self.theta)
        tensor[1, 0] = torch.cos(self.theta) * torch.sin(self.theta)
        tensor[1, 1] = torch.sin(self.theta) ** 2
        tensor[2, 2] = 1
        return tensor


class LeftCircularPolarizer(PolarizedModulationElement):
    r"""
    Left circular polarizer element.

    The left circular polarizer is described by the following polarization matrix:

    .. math::
        J = 
        \begin{bmatrix}
            \frac{1}{2} & -\frac{i}{2} & 0 \\
            \frac{1}{2} & \frac{1}{2} & 0 \\
            0 & 0 & 1
        \end{bmatrix}

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    @property
    def polarized_modulation_profile(self) -> Tensor:
        """Return the polarized modulation profile."""
        tensor = torch.zeros(3, 3, *self.shape, dtype=torch.cdouble, device=next(self.buffers()).device)
        tensor[0, 0] = 0.5
        tensor[0, 1] = -0.5j  # type: ignore
        tensor[1, 0] = 0.5j  # type: ignore
        tensor[1, 1] = 0.5
        tensor[2, 2] = 1
        return tensor


class RightCircularPolarizer(PolarizedModulationElement):
    r"""
    Right circular polarizer element.

    The right circular polarizer is described by the following polarization matrix:

    .. math::
        J =
        \begin{bmatrix}
            \frac{1}{2} & \frac{i}{2} & 0 \\
            \frac{-i}{2} & \frac{1}{2} & 0 \\
            0 & 0 & 1
        \end{bmatrix}
    
    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    @property
    def polarized_modulation_profile(self) -> Tensor:
        """Return the polarized modulation profile."""
        tensor = torch.zeros(3, 3, *self.shape, dtype=torch.cdouble, device=next(self.buffers()).device)
        tensor[0, 0] = 0.5
        tensor[0, 1] = 0.5j  # type: ignore
        tensor[1, 0] = -0.5j  # type: ignore
        tensor[1, 1] = 0.5
        tensor[2, 2] = 1
        return tensor



"""This module defines the IdentityElement element."""

from ..fields import Field
from .elements import Element

__all__ = ["IdentityElement"]


class IdentityElement(Element):
    """
    Identity element.

    The identity element does not modify the input field. It is useful as a placeholder element in
    :class:`torchoptics.System`.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    def forward(self, field: Field) -> Field:
        """Returns the input field without modification."""
        self.validate_field_geometry(field)
        return field
